#import "MyOpenGLView.h"


OBJModel model;

@implementation MyOpenGLView

#pragma mark -
#pragma mark initializations
- (void) heartbeat
{
	[self rotate:standardRotationDegrees];
	if(![[NSApplication sharedApplication] isHidden])
		[self setNeedsDisplay:YES];
	//[self rotateLeft:self];
}

- (void)prepareOpenGL
{
	xTest = 0.0f;
	
	stepSize = 1.0f;
	standardRotationDegrees = 0.3f;
	timer = [NSTimer scheduledTimerWithTimeInterval:(UPDATE_SECONDS) target:self selector:@selector(heartbeat) userInfo:nil repeats:YES];
	[timer retain];
	
	NSLog(@"preparing openGL...");
	glClearColor(1.0f, 0.0f, 0.0f, 1.0f );
	
	//glEnable(GL_COLOR_MATERIAL);
	
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);
	
	yRot = 0.0f;
	xRot = 0.0;;
	upAndDownTranslation = 0.0f;
	leftAndRightTranslation = 0.0f;
	nearAndFarTrasnlation = 0.0f;
	
	//model.loadOBJFile(HIGH_LIGHTCYCLE); /////////////////////////////////////////////////////////
	[self loadModelAtLocation:HIGH_LIGHTCYCLE];
	//GLfloat lightpos[4] = {0.0f, 0.2f, -1.0, 1.0f};
	GLfloat lightpos[4] = {2.0f, 2.0f, 2.0, 1.0f};
	glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	
}

#pragma mark -
#pragma mark Drawing the image
- (void)reshape
{
	[[self openGLContext]makeCurrentContext];
	NSRect bounds = [self bounds];
	//NSLog(@"reshaping: %f, %f", NSWidth(bounds), NSHeight(bounds));
	glViewport(0, 0, NSWidth(bounds), NSHeight(bounds));
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, NSWidth(bounds)/NSHeight(bounds), 0.1f, 100 * model.maxDimension());
	//	glOrtho(-20, 20, -20, 20, -100, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	[NSOpenGLContext clearCurrentContext];
}


- (void)drawRect:(NSRect)rect {
	
	float red, green, blue, alpha;
	[[backgroundColorWell color] getRed:&red green:&green blue:&blue alpha:&alpha];
	glClearColor(red, green, blue, alpha);
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();
	

	glPushMatrix();
	glTranslatef(leftAndRightTranslation, upAndDownTranslation, nearAndFarTrasnlation);
	
	glRotatef(xRot, 1.0f, 0.0f, 0.0f);
	glRotatef(yRot, 0.0f, 0.0f, 1.0f);
	
	//draw the model
	model.drawModel();
	glPopMatrix();
	[[self openGLContext]flushBuffer];
}

#pragma mark -
#pragma mark Moving The Camera
- (void)rotate:(float)degrees
{
	yRot = yRot - degrees;
}
- (IBAction)rotateUp:(id)sender
{
	xRot-= 5.0f;
	if(xRot < 0.1f)
		xRot = 359.9;
	[self setNeedsDisplay:YES];
}
- (IBAction)rotateDown:(id)sender
{
	xRot += 5.0f;
	if(xRot > 359.9f)
		xRot = 0.0f;
	[self setNeedsDisplay:YES];
}
- (IBAction)rotateLeft:(id)sender
{
	yRot -= 5.0f;
	if(yRot < 0.1f)
		yRot = 359.9;
	[self setNeedsDisplay:YES];
}
- (IBAction)rotateRight:(id)sender
{
	yRot += 5.0f;
	if(yRot > 359.9f)
		yRot = 0.0f;
	[self setNeedsDisplay:YES];
}

-(IBAction)stopDefaultRotation:(id)sender
{
	if(standardRotationDegrees > 0.0f)
		standardRotationDegrees = 0.0f;
	else
		standardRotationDegrees = STANDARD_ROTATION_DEGREES_PREDEFINED_VALUE;
}


- (IBAction)moveCameraLeft:(id)sender
{
	leftAndRightTranslation -= stepSize;
	[self setNeedsDisplay:YES];
}
- (IBAction)moveCameraRight:(id)sender
{
	leftAndRightTranslation += stepSize;
	[self setNeedsDisplay:YES];
}
- (IBAction)moveCameraUp:(id)sender
{
	upAndDownTranslation += stepSize;
	[self setNeedsDisplay:YES];
}
- (IBAction)moveCameraDown:(id)sender
{
	upAndDownTranslation -= stepSize;
	[self setNeedsDisplay:YES];
}
- (IBAction)moveCameraForward:(id)sender
{
	nearAndFarTrasnlation += stepSize;
	[self setNeedsDisplay:YES];
}
- (IBAction)moveCameraBackwards:(id)sender
{
	nearAndFarTrasnlation -= stepSize;
	[self setNeedsDisplay:YES];
}

#pragma mark -
#pragma mark Functions
- (void)loadModelAtLocation:(NSString *)location
{
	//model.reset();
	model.loadOBJFile([[location stringByStandardizingPath]cStringUsingEncoding:NSUTF8StringEncoding]);
	[modelFileLocationTextField setStringValue:location];
	
	if (model.hasTextures)
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);
}


- (void)resetModelPosition
{
	upAndDownTranslation = 0.0f;
	leftAndRightTranslation = 0.0f;
	nearAndFarTrasnlation = 0.0f;
	[self setNeedsDisplay:YES];
}

#pragma mark -
#pragma mark IBActions
- (IBAction)loadOBJFile:(id)sender
{
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
	NSArray *allowedFileTypes = [NSArray arrayWithObject:@"obj"];
	[openPanel setCanChooseDirectories:NO];
	[openPanel setCanChooseFiles:YES];
	[openPanel setAllowsMultipleSelection:NO];
	
	[openPanel setAllowedFileTypes:allowedFileTypes];
	if ([openPanel runModalForTypes:allowedFileTypes] == NSOKButton) {
		[self loadModelAtLocation:[openPanel filename]];
	}
}

- (IBAction)resetModelPosition:(id)sender
{
	[self resetModelPosition];
}

- (IBAction)comboBoxSelectionDidChange:(id)sender
{
	NSLog(@"changing step size...");
	NSLog([sender titleOfSelectedItem]);
	if ([[sender titleOfSelectedItem]isEqualToString:@"Other..."]) {
		[stepSizeTextField setEnabled:YES];
		stepSize = [stepSizeTextField floatValue];
	}
	else
		stepSize = [[sender titleOfSelectedItem]floatValue];
	[self setNeedsDisplay:YES];
}


- (IBAction)toggleTextures:(id)sender
{
	GLboolean areTexturesEnabled;
	glGetBooleanv(GL_TEXTURE_2D, &areTexturesEnabled);
	NSLog(@"GL_TEXTURE_2D value is: %i", areTexturesEnabled);
	if (areTexturesEnabled == GL_FALSE)
		glEnable(GL_TEXTURE_2D);
	else
		glDisable(GL_TEXTURE_2D);
	[self setNeedsDisplay:YES];
}

- (IBAction)toggleLighting:(id)sender
{
	GLboolean isLightingEnabled;
	glGetBooleanv(GL_LIGHTING, &isLightingEnabled);
	NSLog(@"GL_LIGHTING value is: %i", isLightingEnabled);
	if (isLightingEnabled == GL_FALSE)
		glEnable(GL_LIGHTING);
	else
		glDisable(GL_LIGHTING);
	[self setNeedsDisplay:YES];
}
#pragma mark -
#pragma mark NSTextField delegations

- (void)controlTextDidEndEditing:(NSNotification *)aNotification
{
	stepSize = [stepSizeTextField floatValue];
	[stepSizeTextField setEnabled:NO];
	[self setNeedsDisplay:YES];
}

#pragma mark -
#pragma mark Mouse Movements
- (void)mouseDown:(NSEvent *)theEvent
{
	
	NSLog(@"mouse down %i", xTest);
}

- (void)mouseDragged:(NSEvent *)theEvent
{
	NSPoint mouse = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	
	NSLog(@"%f", mouse.x - lastMousePoint.x);
	
	if (mouse.x - lastMousePoint.x > 0)
		leftAndRightTranslation += stepSize;
	else
		leftAndRightTranslation -= stepSize;
	if (mouse.y - lastMousePoint.y > 0)
		upAndDownTranslation += stepSize;
	else
		upAndDownTranslation -= stepSize;
	
	lastMousePoint = mouse;
	
	[self setNeedsDisplay:YES];
}

- (void)rightMouseDown:(NSEvent *)event
{
	NSPoint mouse = [self convertPoint:[event locationInWindow] fromView:nil];
	
	NSLog(@"%f", mouse.x - lastMousePoint.x);
	
	if (mouse.x - lastMousePoint.x > 0)
		nearAndFarTrasnlation += stepSize;
	else
		nearAndFarTrasnlation -= stepSize;
	
	lastMousePoint = mouse;
	
	[self setNeedsDisplay:YES];
}

- (void)mouseUp:(NSEvent *)theEvent
{
	NSLog(@"mouse up");
}

#pragma mark -
#pragma mark  deallocations

- (void) dealloc
{
	if(timer != nil)
		[timer release];
	[super dealloc];
}

@end
