#ifndef OBJLOADERS_H
#define OBJLOADERS_H

#include "tgaloaders.h"
#include <iostream>
#include <vector>
#include <string>
#include <OpenGL/OpenGL.h>
#include <OpenGL/glu.h>
#include "MTLLoaders.h"
#include "Routines.hh"
#import <Cocoa/Cocoa.h>
#import "TextureLoaders.h"
#include <SFCocoa/SFCocoa.h>

using namespace std;

//class MTLMaterial;
//typedef float vector3f[3];

class vector3f{
	float zeroFloat;
public:
	float x,y,z;
	vector3f();
	vector3f(float a, float b, float c);
	
	float& operator[](short index);
	
	float magnitude(){return sqrt(x*x + y*y + z*z);}
	void normalize(){
		float magSq = x*x + y*y + z*z;
		if (magSq > 0.0f) {
			float oneOverMag = 1.0f/ sqrt(magSq);
			x *= oneOverMag;
			y *= oneOverMag;
			z *= oneOverMag;
		}
	}
};

class OBJFileInfo{
public:
	unsigned numVertices;
	unsigned numTriangles;
	unsigned numVertexNormals;
	unsigned numTextureVertex;
	unsigned numGroups;
	OBJFileInfo();
};


class Triangle{
public:
	unsigned TriangleEdges[3];			//Triangle Edges Indices (pointing in the main model)
	unsigned TriangleTextureCoords[3];	//Texture Indices (pointing in the main model)
	unsigned TriangleNormals[3];		//Normalized Vectors Indices (pointing in the main model)
	
	MTLMaterial triangleMaterial;
	GLuint textureID;
	//unsigned int triangleMaterial; // triangle material index
	
	void showEdgesIndices(){
		cout << TriangleEdges[0] << " " << TriangleEdges[1] << " " << TriangleEdges[2] << '\n';
	}
	Triangle(const Triangle& otherTriangle);
	Triangle();
	void reset();
};

class OBJGroup{
public:
	string name;
	unsigned numTriangles;
	vector<unsigned> triangleIndicesIndexes;
	GLuint material; //index to material for group;
	
	void showGroupTriangleIndices(){
		for (int i=0; i<triangleIndicesIndexes.size(); i++) {
			cout << triangleIndicesIndexes[i] << '\n';
		}
	}
	
};

class OBJModel{
	int tgaTexturePositionWithTexID(GLuint texID);
	string directoryPrefix;
	
	string findDirectoryPrefix(const char *filename);
public:
	bool hasTextures;
	bool hasNormals;
	
	float xMaxSize;
	float xMinSize;
	float yMaxSize;
	float yMinSize;
	float zMaxSize;
	float zMinSize;
	
	vector<vector3f> vertices;
	vector<vector3f> vertexNormals;
	vector<vector3f> textureVertices;
	
	vector<Triangle> triangles;
	vector<MTLMaterial> materials;
	//vector<TGATexture> textures;
	vector<BitmapTexture> textures;
	vector<string> mtlFileLocations;
	
	//experimental
	vector<OBJGroup> groups;
	//unsigned currentGroupPosition;

	OBJModel(){};
	~OBJModel();
	void allocUsingInfo(OBJFileInfo &info);
	void vertexAtIndex(unsigned index, vector3f resultBuffer);
	void drawVertexAtIndex(unsigned index);
	void reset();
	vector<Triangle> trianglesForGroupAtIndex(unsigned index);
	
	void drawModel();
	//TO BE COMPLETED!!!
	void drawFaceAtIndex(unsigned index);
	void showVertices();
	void showNormals();
	
	float maxDimension();
	float minDimension();
	unsigned numVertices();
	unsigned numTriangles();
	unsigned numVertexNormals();
	unsigned numTextureVertex();
	
	void loadMtlFromFile(const char *filename);
	void loadOBJFile(const char *filename);
	
};

#pragma mark -
void firstPass(const char *filename, OBJFileInfo& info);

#endif