#include "ObjLoaders.h"

#pragma mark -
#pragma mark vector3f
vector3f::vector3f(){x = 0; y = 0; z = 0; zeroFloat = 0.0f;}
vector3f::vector3f(float a, float b, float c){x = a; y = b; z = c; zeroFloat = 0.0f;}

float& vector3f::operator[](short index){
	//if(index > 2 || index < 0)
	//	return 0.0f;
	switch (index) {
		case 0:
			return x;
			break;
		case 1:
			return y;
			break;
		case 2:
			return z;
			break;
		default:
			break;
	}
	return zeroFloat;
}

#pragma mark -
#pragma mark ObjFileInfo

OBJFileInfo::OBJFileInfo(){
	numVertices = 0;
	numTriangles = 0;
	numVertexNormals = 0;
	numTextureVertex = 0;
	numGroups = 0;
}


#pragma mark -
#pragma mark Triangle

Triangle::Triangle(const Triangle& otherTriangle){
	for (int i=0; i<3; i++) {
		TriangleEdges[i] = otherTriangle.TriangleEdges[i];
		TriangleTextureCoords[i] = otherTriangle.TriangleTextureCoords[i];
		TriangleNormals[i] = otherTriangle.TriangleNormals[i];
	}
	triangleMaterial = otherTriangle.triangleMaterial;
	textureID = otherTriangle.textureID;
}
Triangle::Triangle(){
	reset();
}
void Triangle::reset(){
	for (int i=0; i<3; i++) {
		TriangleEdges[i] = 0;
		TriangleTextureCoords[i] = 0;
		TriangleNormals[i] = 0;
	}
	textureID = -1;
}


#pragma mark -
#pragma mark OBJModel

string OBJModel::findDirectoryPrefix(const char *filename){
	string path = filename;
	unsigned loc = path.find_last_of('/');
	if(loc == string::npos)
		return string("");
	else
		return path.substr(0,path.length() - (path.length()-loc-1));
}

int OBJModel::tgaTexturePositionWithTexID(GLuint texID){
	for (int i=0; i<textures.size(); i++) {
		//if(textures[i].texture.texID == texID) return i;
		if(textures[i].texID == texID) return i;
	}
	return -1;
}

void OBJModel::vertexAtIndex(unsigned index, vector3f resultBuffer){
	resultBuffer[0] = vertices[index][0];
	resultBuffer[1] = vertices[index][1];
	resultBuffer[2] = vertices[index][2];
}

void OBJModel::drawVertexAtIndex(unsigned index){
	vector3f vertexToDraw;
	vertexToDraw[0] = vertices[index][0];
	vertexToDraw[1] = vertices[index][1];
	vertexToDraw[2] = vertices[index][2];
	
	glVertex3f(vertexToDraw[0], vertexToDraw[1], vertexToDraw[2]);
	
}

vector<Triangle> OBJModel::trianglesForGroupAtIndex(unsigned index){
	vector<Triangle> result;
	for (int i=0; i<groups[index].triangleIndicesIndexes.size(); i++)
		result.push_back(triangles[groups[index].triangleIndicesIndexes[i]]);
	return result;
}

void OBJModel::drawModel()
{
	
	for (int i=0; i<this->triangles.size(); i++) {
		this->drawFaceAtIndex(i);
	}
}

//TO BE COMPLETED!!!
void OBJModel::drawFaceAtIndex(unsigned index){
	static int lastTexUsed = -1;
	//int currentTexturePosition = this->tgaTexturePositionWithTexID(triangles[index].textureID);
	
	unsigned triangleVertexPositions[3];
	triangleVertexPositions[0] = triangles[index].TriangleEdges[0];
	triangleVertexPositions[1] = triangles[index].TriangleEdges[1];
	triangleVertexPositions[2] = triangles[index].TriangleEdges[2];
	
	unsigned triangleNormalPositions[3];
	triangleNormalPositions[0] = triangles[index].TriangleNormals[0];
	triangleNormalPositions[1] = triangles[index].TriangleNormals[1];
	triangleNormalPositions[2] = triangles[index].TriangleNormals[2];
	
	unsigned vertexTexturePositions[3];
	vertexTexturePositions[0] = triangles[index].TriangleTextureCoords[0];
	vertexTexturePositions[1] = triangles[index].TriangleTextureCoords[1];
	vertexTexturePositions[2] = triangles[index].TriangleTextureCoords[2];
	
	//cout << triangles[index].TriangleEdges[0] << " " << triangles[index].TriangleEdges[0] << " " << triangles[index].TriangleEdges[0] << '\n';
	
	vector3f v1, v2, v3;
	vector3f n1, n2, n3;
	vector3f t1, t2, t3;
	
	for(unsigned i=0; i<3; ++i){
		v1[i] = vertices[triangleVertexPositions[0]][i];
		v2[i] = vertices[triangleVertexPositions[1]][i];
		v3[i] = vertices[triangleVertexPositions[2]][i];
		
		n1[i] = vertexNormals[triangleNormalPositions[0]][i];
		n2[i] = vertexNormals[triangleNormalPositions[1]][i];
		n3[i] = vertexNormals[triangleNormalPositions[2]][i];
		
		if(this->hasTextures){
			t1[i] = textureVertices[vertexTexturePositions[0]][i];
			t2[i] = textureVertices[vertexTexturePositions[1]][i];
			t3[i] = textureVertices[vertexTexturePositions[2]][i];
		}
	}		
	
	//n1.normalize();
	//n2.normalize();
	//n3.normalize();
	
	MTLMaterial currentMaterial;
	currentMaterial = triangles[index].triangleMaterial;
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, currentMaterial.ambient);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, currentMaterial.diffuse);
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, currentMaterial.specular);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, currentMaterial.shininess);
	
	if(this->hasTextures && lastTexUsed != this->triangles[index].textureID){
		//glEnable(GL_TEXTURE_2D);
		
		if(this->triangles[index].textureID == -1)
			glBindTexture(GL_TEXTURE_2D, 0);
		else
			glBindTexture(GL_TEXTURE_2D, this->triangles[index].textureID);
		
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		
		lastTexUsed = this->triangles[index].textureID;
	}
	
	glBegin(GL_TRIANGLES);
	
	glNormal3f(n1[0],n1[1],n1[2]);
	if (this->hasTextures) glTexCoord2f(t1[0], t1[1]);
	glVertex3f(v1[0],v1[1],v1[2]);
	
	glNormal3f(n2[0],n2[1],n2[2]);
	if (this->hasTextures) glTexCoord2f(t2[0], t2[1]);
	glVertex3f(v2[0],v2[1],v2[2]);
	
	glNormal3f(n3[0],n3[1],n3[2]);
	if (this->hasTextures) glTexCoord2f(t3[0], t3[1]);
	glVertex3f(v3[0],v3[1],v3[2]);
	
	glEnd();
}

void OBJModel::reset(){
	NSLog(@"resetting model...");
	//for (int i=0; i<textures.size(); i++) {
	//	glDeleteTextures(1, &this->textures[i].texture.texID);
	//}
	this->vertices.clear();
	this->vertexNormals.clear();
	this->textureVertices.clear();
	this->triangles.clear();
	this->mtlFileLocations.clear();
	this->materials.clear();
	this->groups.clear();
	this->textures.clear();
	
	hasTextures = false;
	hasNormals = false;
	
	xMaxSize = 0.0f;
	xMinSize = 0.0f;
	yMaxSize = 0.0f;
	yMinSize = 0.0f;
	zMaxSize = 0.0f;
	zMinSize = 0.0f;
	directoryPrefix = "";
}

OBJModel::~OBJModel(){}

void OBJModel::showVertices()
{
	for (int i=0; i<vertices.size(); i++) 
		cout << vertices[i][0] << " " << vertices[i][1] << " " << vertices[i][2] << '\n';
	cout << '\n';
}

void OBJModel::showNormals()
{
	for (int i=0; i<vertexNormals.size(); i++)
		cout << vertexNormals[i][0] << " " << vertexNormals[i][1] << " " << vertexNormals[i][2] << '\n';
	cout << '\n';
}

float OBJModel::maxDimension()
{
	float result = this->xMaxSize;
	if(result > this->yMaxSize) result = this->yMaxSize;
	if(result > this->zMaxSize) result = this->zMaxSize;
	return result;
}

float OBJModel::minDimension()
{
	float result = this->xMinSize;
	if(result < this->yMinSize) result = this->yMinSize;
	if(result < this->zMinSize) result = this->zMinSize;
	return result;
}

unsigned OBJModel::numVertices(){return vertices.size();}
unsigned OBJModel::numTriangles(){return triangles.size();}
unsigned OBJModel::numVertexNormals(){return vertexNormals.size();}
unsigned OBJModel::numTextureVertex(){return textureVertices.size();}

void OBJModel::loadOBJFile(const char *filename){
	this->reset();
	
	OBJFileInfo currentFileInfo;
	
	directoryPrefix = this->findDirectoryPrefix(filename);
	cout << "directoryPrefix is: " << this->directoryPrefix << '\n';
	firstPass(filename, currentFileInfo);
	groups.assign(currentFileInfo.numGroups, OBJGroup());
	
	MTLMaterial currentMaterial; //set the current material
	
	char buf[128];
	FILE *filePointer = fopen(filename, "r");
	
	if (!filePointer) {
		cout << "Specified File does not exist!\n";
		return;
	}
	
	unsigned currentVertexIndex = 0;
	unsigned currentvertexNormalIndex = 0;
	unsigned currentTextureVertexIndex = 0;
	unsigned currentTriangleIndex = 0;
	unsigned currentObjGroupIndex = 0;
	bool shouldCurrentObjGroupIndexIncrease = false;
	while ( fscanf(filePointer, "%s", buf) != EOF ) {
		
		switch (buf[0]) {
			case 'm':{
				cout << "mtl lib found:";
				fgets(buf, sizeof(buf), filePointer);	//get the name of the mtl library to load
                sscanf(buf, "%s %s", buf, buf);			//and fix the string
				string newMTLFileLocation =  buf;
				this->mtlFileLocations.push_back(newMTLFileLocation);
				cout << newMTLFileLocation << '\n';
				this->loadMtlFromFile(newMTLFileLocation.c_str()); // load the materials
			}
				break;
			case 'v':{
				vector3f vectorToAdd;
				switch (buf[1]) {
					case '\0':{
						float a,b,c; //(x,y,z)
						
						//try to set the model's boundaries
						if(a > this->xMaxSize) this->xMaxSize = a;
						if(b > this->yMaxSize) this->yMaxSize = b;
						if(c > this->zMaxSize) this->zMaxSize = c;
						if(a < this->xMinSize) this->xMinSize = a;
						if(b < this->yMinSize) this->yMinSize = b;
						if(c < this->zMinSize) this->zMinSize = c;
						
						//add the vertex into the vector created just for vertices
						fscanf(filePointer, "%f %f %f", &a, &b, &c);						
						vectorToAdd[0] = a;
						vectorToAdd[1] = b;
						vectorToAdd[2] = c;
						currentVertexIndex++;
						this->vertices.push_back(vectorToAdd);
					}
						break;
					case 'n':{
						if (this->hasNormals != true) this->hasNormals = true;
						float a,b,c;
						fscanf(filePointer, "%f %f %f", &a, &b, &c);
						vectorToAdd[0] = a;
						vectorToAdd[1] = b;
						vectorToAdd[2] = c;
						this->vertexNormals.push_back(vectorToAdd);
						currentvertexNormalIndex++;
					}
						break;
					case 't':{
						if (this->hasTextures != true) this->hasTextures = true;
						float a,b;	//There is no C. The last element of the vector actually stays '0'
						//It will be handled appropriately when the time for displaying comes...
						fscanf(filePointer, "%f %f", &a, &b);
						vectorToAdd[0] = a;
						vectorToAdd[1] = b;
						this->textureVertices.push_back(vectorToAdd);
						currentTextureVertexIndex++;
					}
						break;
					default:
						break;
				}
			}
				break;
			case 'f':{ // make a triangle
				Triangle currentTriangle;
				int v,t,n = 0;
				fscanf(filePointer, "%s", buf);
				if ( strstr(buf, "//") ) {
					sscanf(buf, "%d//%d", &v, &n);
					//printf("%d %d\n", v, n);
					--v; --n;
					currentTriangle.TriangleEdges[0] = v;
					currentTriangle.TriangleNormals[0] = n;
					fscanf(filePointer, "%d//%d", &v, &n);
					//printf("%d %d\n", v, n);
					--v; --n;
					currentTriangle.TriangleEdges[1] = v;
					currentTriangle.TriangleNormals[1] = n;
					fscanf(filePointer, "%d//%d", &v, &n);
					//printf("%d %d\n", v, n);
					--v; --n;
					currentTriangle.TriangleEdges[2] = v;
					currentTriangle.TriangleNormals[2] = n;
				}
				else if (sscanf(buf, "%d/%d/%d", &v, &t, &n) == 3){
					currentTriangle.TriangleEdges[0] = v;
					currentTriangle.TriangleTextureCoords[0] = t;
					currentTriangle.TriangleNormals[0] = n;
					fscanf(filePointer, "%d/%d/%d", &currentTriangle.TriangleEdges[1], &currentTriangle.TriangleTextureCoords[1], &currentTriangle.TriangleNormals[1]);
					fscanf(filePointer, "%d/%d/%d", &currentTriangle.TriangleEdges[2], &currentTriangle.TriangleTextureCoords[2], &currentTriangle.TriangleNormals[2]);
					for(unsigned i=0; i<3; ++i){
						--currentTriangle.TriangleEdges[i];
						--currentTriangle.TriangleTextureCoords[i];
						--currentTriangle.TriangleNormals[i];
					}
				}
				else if (sscanf(buf, "%d/%d", &v, &t) == 2){
					currentTriangle.TriangleTextureCoords[0] = t;
					currentTriangle.TriangleEdges[0] = v;
					fscanf(filePointer, "%d/%d", &currentTriangle.TriangleEdges[1], &currentTriangle.TriangleTextureCoords[1]);
					fscanf(filePointer, "%d/%d", &currentTriangle.TriangleEdges[2], &currentTriangle.TriangleTextureCoords[2]);
				}
				else {
					sscanf(buf, "%d", &v);
					currentTriangle.TriangleEdges[0] = v;
					fscanf(filePointer, "%d", &currentTriangle.TriangleEdges[1]);
					fscanf(filePointer, "%d", &currentTriangle.TriangleEdges[2]);
				}
				
				groups[currentObjGroupIndex].triangleIndicesIndexes.push_back(currentTriangleIndex);
				
				currentTriangle.triangleMaterial = currentMaterial;

				//cout << this->textures.size();
				if(this->textures.size() == 0){
					currentTriangle.textureID = -1;
				}
				else{
					//currentTriangle.textureID = this->textures[textures.size()-1].texture.texID;
					currentTriangle.textureID = this->textures[textures.size()-1].texID;
					//cout << currentTriangle.textureID;
				}
				
				this->triangles.push_back(currentTriangle);
				currentTriangleIndex++;				
			}
				break;	
				//TO BE FIXED!
			case 'g':{
				if ( (currentObjGroupIndex == 0) && (shouldCurrentObjGroupIndexIncrease == false) ) {
					fgets(buf, sizeof(buf), filePointer);
					groups[currentObjGroupIndex].name = buf;
					shouldCurrentObjGroupIndexIncrease = true;
				}
				else{
					fgets(buf, sizeof(buf), filePointer);
					currentObjGroupIndex++;
					groups[currentObjGroupIndex].name = buf;
				}
			}
				break;
			case 'u':{    //use material
				fgets(buf, sizeof(buf), filePointer);
				sscanf(buf, "%s", buf);
				string nameOfMaterial = buf;
				//cout << "now using material: " << nameOfMaterial << '\n';
				for (int i=0; i < materials.size(); i++) {
					if(materials[i].name == nameOfMaterial){
						//cout << "this exists in library of materials\n";
						currentMaterial = materials[i];
					}
				}
			}
				break;
			default:
				fgets(buf, sizeof(buf), filePointer);
				break;
		}
	}
	fclose(filePointer);
	
	//perform normalization before the display to avoid unnecessary calculations...
	for (int i=0; i<this->vertexNormals.size(); i++) {
		vertexNormals[i].normalize();
	}
}


void OBJModel::loadMtlFromFile(const char *filename){
	string entireFileName = directoryPrefix + filename;
	cout << "Attempting to load MTL file named: " << entireFileName << '\n';
	
	bool shouldIncreaseMaterials = false;
	unsigned currentMTLMaterialIndex = 0;
	
	
	FILE* filepointer = fopen(entireFileName.c_str(), "r");
	
	if (!filepointer) {
		cout << "could not find the MTL file requested. returning function...\n";
		return;
	}
	
	MTLMaterial currentMaterial;
	materials.push_back(currentMaterial);
	
	char buf[128];
	while(fscanf(filepointer, "%s", buf) != EOF){
		
		switch (buf[0]) {
			case '#':
				fgets(buf, sizeof(buf), filepointer);
				break;
			case 'n':{
				fgets(buf, sizeof(buf), filepointer);
				sscanf(buf, "%s", buf);
				if (shouldIncreaseMaterials == false) {
					//currentMaterialRef.name = buf;
					materials[currentMTLMaterialIndex].name = buf;
					shouldIncreaseMaterials = true;
				}
				else{
					MTLMaterial newMaterial;
					newMaterial.name = buf;
					
					materials.push_back(newMaterial);
					currentMTLMaterialIndex++;
					//currentMaterialRef = materials[currentMTLMaterialIndex];
					materials[currentMTLMaterialIndex].name = buf;
				}
				break;
			case 'N':{
				if (buf[1]!='s') break;
				float newShininess = 0.0f;
				fscanf(filepointer, "%f", &newShininess);
				newShininess /= 1000.0f;
				newShininess *= 128.0f;
				materials[currentMTLMaterialIndex].shininess = newShininess;
			}
				break;
			case 'K':
				switch (buf[1]) {
					case 'd':
						fscanf(filepointer, "%f %f %f", &materials[currentMTLMaterialIndex].diffuse[0], &materials[currentMTLMaterialIndex].diffuse[1], &materials[currentMTLMaterialIndex].diffuse[2]);
						break;
					case 's':
						fscanf(filepointer, "%f %f %f", &materials[currentMTLMaterialIndex].specular[0], &materials[currentMTLMaterialIndex].specular[1], &materials[currentMTLMaterialIndex].specular[2]);
						break;
					case 'a':
						fscanf(filepointer, "%f %f %f", &materials[currentMTLMaterialIndex].ambient[0], &materials[currentMTLMaterialIndex].ambient[1], &materials[currentMTLMaterialIndex].ambient[2]);
						break;
					default:
						break;
				}
				break;
			}
				break;
			case 'm':{
				char *textureFileName = (char *)malloc(FILENAME_MAX);
				fgets(textureFileName, FILENAME_MAX, filepointer);
				sscanf(textureFileName, "%s", textureFileName, textureFileName);
				//string textureFileNameStringed = textureFileName;
				//free(textureFileName);				
				if(strncmp(buf, "map_Kd", 6) == 0) 
				{
					string entireTextureFileName = directoryPrefix + textureFileName;
					cout << "texture found: " << entireTextureFileName.c_str() << '\n';
					NSString *entireTextureFileNameNsstring = [NSString stringWithCString:entireTextureFileName.c_str() encoding:NSUTF8StringEncoding];
					/*TGALoaderCPP alpha;
					alpha.LoadTGA(entireTextureFileName.c_str());
					//glEnable(GL_TEXTURE_2D);
					glGenTextures(1, &(alpha.texture.texID));
					glBindTexture(GL_TEXTURE_2D, alpha.texture.texID);
					//glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, alpha.texture.width, alpha.texture.height, 0, alpha.texture.type, GL_UNSIGNED_BYTE, alpha.texture.imageData);
					
					glTexImage2D(GL_TEXTURE_2D, 0, alpha.texture.type, alpha.texture.width, alpha.texture.height, 0, alpha.texture.type, GL_UNSIGNED_BYTE, alpha.texture.imageData);
					glBindTexture(GL_TEXTURE_2D, 0);
					printf("texture id %d\n", alpha.texture.texID);
					textures.push_back(alpha);*/
					
					BitmapTexture alpha;
					alpha.loadImage(entireTextureFileNameNsstring);
					textures.push_back(alpha);
				}
				free(textureFileName);
			}
				break;
			default:
				fgets(buf, sizeof(buf), filepointer);
				break;
		}
	}
	fclose(filepointer);
	
}


#pragma mark -
#pragma mark Solo Functions

void firstPass(const char *filename, OBJFileInfo& info)
{
	char buf[128];
	FILE *filePointer = fopen(filename, "r");	
	
	while ( fscanf(filePointer, "%s", buf) != EOF){
		switch (buf[0]) {
			case 'v':
				switch (buf[1]) {
					case 't':
						fgets(buf, sizeof(buf), filePointer);
						info.numTextureVertex++;
						break;
					case '\0':
						fgets(buf, sizeof(buf), filePointer);
						info.numVertices++;
						break;
					case 'n':
						fgets(buf, sizeof(buf), filePointer);
						info.numVertexNormals++;
						break;
					default:
						break;
				}
				break;
			case '#':
				fgets(buf, sizeof(buf), filePointer);
				break;
			case 'f':
				info.numTriangles++;
				//NSLog(@"not implemented yet! %i", info.numTriangles);
				break;
			case 'g':
				info.numGroups++;
				break;
			default:
				fgets(buf, sizeof(buf), filePointer);
				break;
		}
	}
	
	if(info.numGroups == 0)		//if the file does not issue the group command, let's at least
		info.numGroups++;		//explicitly define the default group
	
	fclose(filePointer);
	info.numTriangles;//--;
}

